#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <wait.h>

#define ENTREE 0
#define SORTIE 1

#define UINT unsigned int

typedef struct couple
{
  UINT first,second;
} Couple;

/*
 * (1) Calculer le temps d'excution du programme (on nglige le temps
 * d'excution des instructions autres que <<sleep>>).
 * Comment l'amliorer ? (sans supprimer l'instruction sleep !)
 *
 * (2) Ecrire une version avec seulement deux tubes.
 */

void swap(UINT *a,UINT *b)
{
  UINT aux;
  
  aux=*a;
  *a = *b;
  *b = aux;
}

void envoyer(int fd, Couple  *p)
{
	write(fd, p, sizeof(Couple)); 
}

int recevoir(int fd, Couple *p)
{
	return read(fd, p, sizeof(Couple));
}

void fils(int entree, int sortie)
{
  Couple p;

  while ( recevoir(entree, &p) ) /* tant que le pre a besoin du fils */
    {
      printf("Le fils [%d] reoit %d puis %d.\n", 
	     getpid(),p.first,p.second);
      sleep(2);
      if ( p.first > p.second) 
	swap(&p.first,&p.second);
      envoyer(sortie, &p);
    }
  printf("Fin du fils [%d].\n", getpid());
  exit(EXIT_SUCCESS);
}

int main(void)
{
  /* p = pre, a = premier fils, b = second fils
   *
   * pa = tube pre -> premier fils
   * ap = tube pre <- premier fils
   * pb = tube pre -> second fils
   * bp = tube pre <- second fils
   */
  
  int pa[2];
  int ap[2];
  int pb[2];
  int bp[2];
  Couple ab,cd;
	
  printf("Rentrer a/b/c/d :");
  if(scanf("%u/%u/%u/%u",
	   &ab.first,&ab.second,&cd.first,&cd.second) != 4) 
    exit(EXIT_FAILURE);
  
  pipe(pa);
  pipe(ap);
  pipe(pb);
  pipe(bp);
  
  if ( fork() ) /* pre */
    {
      if ( fork() ) /* pre */
	{
	  close(pa[ENTREE]); close(ap[SORTIE]);
	  close(pb[ENTREE]); close(bp[SORTIE]);
	  
	  /* Premire tape */
	  envoyer(pa[SORTIE], &ab);
	  recevoir(ap[ENTREE], &ab);
	  envoyer(pb[SORTIE], &cd);
	  recevoir(bp[ENTREE], &cd);

	  /* Deuxime tape */
	  swap(&ab.second,&cd.first);
	  envoyer(pa[SORTIE], &ab);
	  recevoir(ap[ENTREE], &ab);
	  envoyer(pb[SORTIE], &cd );
	  recevoir(bp[ENTREE], &cd);

	  /* Troisime tape */
	  swap(&ab.second,&cd.second);
	  swap(&cd.first,&cd.second);
	  envoyer(pa[SORTIE], &cd );
	  recevoir(ap[ENTREE], &cd);
	  swap(&cd.first,&cd.second);
	  swap(&ab.second,&cd.second);

	  
	  printf("Rsultat du tri : %d %d %d %d.\n", 
		 ab.first,ab.second,
		 cd.first,cd.second);
	  close(pa[SORTIE]); close(ap[ENTREE]);
	  close(pb[SORTIE]); close(bp[ENTREE]);
	  wait(NULL);
	  wait(NULL);
	}
      else /* second fils cr */
	{
	  close(pa[ENTREE]); close(pa[SORTIE]);
	  close(ap[ENTREE]); close(ap[SORTIE]);
	  close(pb[SORTIE]); close(bp[ENTREE]);
	  fils(pb[ENTREE], bp[SORTIE]);
	}			
    }
  else /* premier fils cr */
    {
      close(pb[ENTREE]); close(pb[SORTIE]);
      close(bp[ENTREE]); close(bp[SORTIE]);
      close(pa[SORTIE]); close(ap[ENTREE]);
      fils(pa[ENTREE], ap[SORTIE]);
    }
  return EXIT_SUCCESS;
}



